# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import socket

import httplib2
import json

from copy import deepcopy

from collections import defaultdict

from oslo_config import cfg

from oslo_log import log as logging

from oslo_concurrency import lockutils

from oslo_service import loopingcall

from networking_bagpipe._i18n import _

from networking_bagpipe.agent.common import constants as b_const

from neutron.conf.agent import common as config

from neutron_lib import exceptions as n_exc

LOG = logging.getLogger(__name__)


# Having this at line 231 is apparently not enough, so adding here as well:
# pylint: disable=not-callable

bagpipe_bgp_opts = [
    cfg.IntOpt('ping_interval', default=10,
               help=_("The number of seconds the bagpipe-bgp client will "
                      "wait between polling for restart detection.")),
    cfg.PortOpt('bagpipe_bgp_port', default=8082,
                help=_("bagpipe-bgp REST service IP port.")),
]

# these options are for internal use only (fullstack tests), and hence
# better kept in a separate table not looked at by oslo gen confi hooks
internal_opts = [
    cfg.HostAddressOpt('bagpipe_bgp_ip', default='127.0.0.1',
                       help=_("bagpipe-bgp REST service IP address.")),
]

cfg.CONF.register_opts(bagpipe_bgp_opts, "BAGPIPE")
cfg.CONF.register_opts(internal_opts, "BAGPIPE")
config.register_agent_state_opts_helper(cfg.CONF)


class BaGPipeBGPException(n_exc.NeutronException):
    message = "An exception occurred when calling bagpipe-bgp \
               REST service: %(reason)s"


class HTTPClientBase(object):
    """An HTTP client base class"""

    def __init__(self, host="127.0.0.1", port=8082,
                 client_name="HTTP client base"):
        """Create a new HTTP client

        :param host: HTTP server IP address
        :param port: HTTP server port
        """
        self.host = host
        self.port = port
        self.client_name = client_name

    def do_request(self, method, action, body=None):
        LOG.debug("bagpipe-bgp client request: %(method)s %(action)s "
                  "[%(body)s]",
                  {'method': method, 'action': action, 'body': str(body)})

        if isinstance(body, dict):
            body = json.dumps(body)
        try:
            headers = {'User-Agent': self.client_name,
                       "Content-Type": "application/json",
                       "Accept": "application/json"}
            uri = "http://%s:%s/%s" % (self.host, self.port, action)

            http = httplib2.Http()
            response, content = http.request(uri, method, body, headers)
            LOG.debug("bagpipe-bgp returns [%(status)s:%(content)s]",
                      {'status': str(response.status), 'content': content})

            if response.status == 200:
                if content and len(content) > 1:
                    return json.loads(content)
            else:
                reason = (
                    "An HTTP operation has failed on bagpipe-bgp."
                )
                raise BaGPipeBGPException(reason=reason)
        except (socket.error, IOError) as e:
            reason = "Failed to connect to bagpipe-bgp: %s" % str(e)
            raise BaGPipeBGPException(reason=reason)

    def get(self, action):
        return self.do_request("GET", action)

    def post(self, action, body=None):
        return self.do_request("POST", action, body=body)

    def put(self, action, body=None):
        return self.do_request("PUT", action, body=body)

    def delete(self, action):
        return self.do_request("DELETE", action)


class BaGPipeBGPAgent(HTTPClientBase):
    """Implements a BaGPipe-BGP REST client"""

    _instance = None

    # bagpipe-bgp status
    BAGPIPEBGP_UP = 'UP'
    BAGPIPEBGP_DOWN = 'DOWN'

    @classmethod
    @lockutils.synchronized('bagpipe-bgp-agent')
    def _create_instance(cls, agent_type):
        if not cls.has_instance():
            cls._instance = cls(agent_type)

    @classmethod
    def has_instance(cls):
        return cls._instance is not None

    @classmethod
    def clear_instance(cls):
        cls._instance = None

    @classmethod
    def get_instance(cls, agent_type):
        # double checked locking
        if not cls.has_instance():
            cls._create_instance(agent_type)
        else:
            if cls._instance.agent_type != agent_type:
                raise Exception("Agent already configured with another type")

        return cls._instance

    def __init__(self, agent_type):

        """Create a new BaGPipe-BGP REST service client.

        :param agent_type: bagpipe-bgp agent type (Linux bridge or OVS)

        """
        super(BaGPipeBGPAgent,
              self).__init__(cfg.CONF.BAGPIPE.bagpipe_bgp_ip,
                             cfg.CONF.BAGPIPE.bagpipe_bgp_port, agent_type)

        self.agent_type = agent_type

        self.ping_interval = cfg.CONF.BAGPIPE.ping_interval

        self.bagpipe_bgp_status = self.BAGPIPEBGP_DOWN
        self.seq_num = 0

        # Starts a greenthread for bagpipe-bgp status polling
        self._start_bagpipe_bgp_status_polling(self.ping_interval)

        # Maps registered service name to callback function used to build the
        # content of bagpipe-bgp API calls
        self.build_callbacks = dict()

        # Maps registered service name to list of ports for which a given
        # service is enabled
        self.port_lists = dict()

    def _check_bagpipe_bgp_status(self):
        """Trigger refresh on bagpipe-bgp restarts

        Check if bagpipe-bgp has restarted while sending ping request
        to detect sequence number change.
        If a restart is detected, re-send all registered attachments to
        bagpipe-bgp.
        """
        new_seq_num = self._request_ping()

        # Check bagpipe-bgp restart
        if new_seq_num != self.seq_num:
            if new_seq_num != -1:
                if self.seq_num != 0:
                    LOG.warning("bagpipe-bgp restart detected...")
                else:
                    LOG.info("bagpipe-bgp successfully detected")

                self.seq_num = new_seq_num
                self.bagpipe_bgp_status = self.BAGPIPEBGP_UP

                # Re-send all registered ports to bagpipe-bgp
                ports = set()
                for port_list in self.port_lists.values():
                    ports |= port_list

                if ports:
                    LOG.info("Sending all registered ports to bagpipe-bgp")
                    LOG.debug("Registered ports list: %s", ports)
                    for port_id in ports:
                        self.do_port_plug(port_id)
                else:
                    LOG.info("No attachment to send to bagpipe-bgp")
            else:
                self.bagpipe_bgp_status = self.BAGPIPEBGP_DOWN

    def _start_bagpipe_bgp_status_polling(self, ping_interval=10):
        # Start bagpipe-bgp status polling at regular interval
        status_loop = loopingcall.FixedIntervalLoopingCall(
            self._check_bagpipe_bgp_status)
        status_loop.start(interval=ping_interval,
                          initial_delay=ping_interval)  # TM: why not zero ?

    def _check_evpn2ipvpn_info(self, vpn_type, network_id, attach_list,
                               attach_info):
        # Check if plugging an EVPN into an IPVPN
        if (vpn_type == b_const.IPVPN and b_const.EVPN in attach_list):
            attach_info['local_port'] = {
                b_const.EVPN: {
                    'id': '%s_evpn' % network_id
                }
            }

    def _compile_port_attach_info(self, port_id):
        service_attach_info = {}
        for service, build_callback in self.build_callbacks.items():
            service_attach_info[service] = build_callback(port_id)

        attach_list = defaultdict(list)
        for vpn_type in b_const.VPN_TYPES:
            attach_info = {}

            for service in self.build_callbacks.keys():
                if vpn_type in service_attach_info[service]:
                    network_id = service_attach_info[service]['network_id']
                    service_info = service_attach_info[service]
                    if not attach_info:
                        attach_info = dict(
                            vpn_instance_id='%s_%s' % (network_id, vpn_type),
                            ip_address=service_info['ip_address'],
                            mac_address=service_info['mac_address'],
                            gateway_ip=service_info['gateway_ip'],
                            local_port=service_info['local_port']
                        )

                    service_vpn_info = service_info[vpn_type]

                    if vpn_type not in attach_info:
                        attach_info.update(dict(vpn_type=vpn_type))

                    attach_info['local_port'].update(
                        service_vpn_info.pop('local_port', {})
                    )

                    self._check_evpn2ipvpn_info(vpn_type, network_id,
                                                attach_list, attach_info)

                    # Check if static routes
                    static_routes = service_vpn_info.pop('static_routes', [])
                    if static_routes:
                        static_info = deepcopy(attach_info)
                        static_info.update({'advertise_subnet': True})
                        for static_route in static_routes:
                            static_info['ip_address'] = static_route
                            static_info.update(service_vpn_info)

                            attach_list[vpn_type].append(static_info)

                    for rt_type in b_const.RT_TYPES:
                        if rt_type in service_vpn_info:
                            if rt_type not in attach_info:
                                attach_info[rt_type] = []

                            attach_info[rt_type] += (
                                service_vpn_info.pop(rt_type)
                            )

                    attach_info.update(service_vpn_info)

            if attach_info:
                attach_list[vpn_type].append(attach_info)

        return attach_list

    def _request_ping(self):
        """Send ping request to bagpipe-bgp to get sequence number"""
        try:
            response = self.get('ping')
            LOG.debug("bagpipe-bgp PING response received with "
                      "sequence number %s", response)
            return response
        except BaGPipeBGPException as e:
            LOG.warning(str(e))
            return -1

    def _send_attach_local_port(self, local_port_details):
        """Send local port attach request to BaGPipe-BGP if running"""
        if self.bagpipe_bgp_status is self.BAGPIPEBGP_UP:
            try:
                self.post('attach_localport', local_port_details)
                LOG.debug("Local port has been attached to bagpipe-bgp with "
                          "details %s", local_port_details)
            except BaGPipeBGPException as e:
                LOG.error("Can't attach local port on bagpipe-bgp: %s", str(e))
        else:
            LOG.debug("Local port not yet attached to bagpipe-bgp (not up)")

    def _send_detach_local_port(self, local_port_details):
        """Send local port detach request to BaGPipe-BGP if running"""
        if self.bagpipe_bgp_status is self.BAGPIPEBGP_UP:
            try:
                self.post('detach_localport', local_port_details)
                LOG.debug("Local port has been detached from bagpipe-bgp "
                          "with details %s", local_port_details)
            except BaGPipeBGPException as e:
                LOG.error("Can't detach local port from bagpipe-bgp: %s",
                          str(e))
                raise
        else:
            LOG.debug("Local port not yet detached from bagpipe-bgp (not up)")

    def do_port_plug(self, port_id):
        """Send port attach request to bagpipe-bgp."""
        all_plug_details = self._compile_port_attach_info(port_id)

        # First plug E-VPNs because they could be plugged into IP-VPNs
        for vpn_type in [t for t in b_const.VPN_TYPES
                         if t in all_plug_details]:
            for plug_detail in all_plug_details[vpn_type]:
                self._send_attach_local_port(plug_detail)

    def do_port_plug_refresh(self, port_id, detach_infos):
        """Refresh port attach on bagpipe-bgp

        Send port attach and/or detach request to bagpipe-bgp when necessary.
        """
        plug_details = self._compile_port_attach_info(port_id)

        for detach_vpn_type, detach_info in list(detach_infos.items()):
            if detach_vpn_type not in plug_details.keys():
                network_id = detach_info.pop('network_id')
                detach_info.update({'vpn_instance_id': '%s_%s' %
                                    (network_id, detach_vpn_type),
                                    'vpn_type': detach_vpn_type})

                self._check_evpn2ipvpn_info(detach_vpn_type, network_id,
                                            plug_details, detach_info)
            else:
                del detach_infos[detach_vpn_type]

        if detach_infos:
            for vpn_type in [t for t in b_const.VPN_TYPES[::-1]
                             if t in detach_infos]:
                self._send_detach_local_port(detach_infos[vpn_type])

        for vpn_type in [t for t in b_const.VPN_TYPES
                         if t in plug_details]:
            for plug_detail in plug_details[vpn_type]:
                self._send_attach_local_port(plug_detail)

    @lockutils.synchronized('bagpipe-bgp-agent')
    def register_build_callback(self, service_name, callback):
        self.build_callbacks[service_name] = callback

    @lockutils.synchronized('bagpipe-bgp-agent')
    def register_port_list(self, service_name, port_list):
        self.port_lists[service_name] = port_list



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from collections import defaultdict

from networking_bagpipe.agent.common import constants as b_const

LOG = logging.getLogger(__name__)


class keydefaultdict(defaultdict):
    """Inherit defaultdict class to customize default_factory.

    Override __missing__ method to construct custom object as default_factory
    passing an argument.

    For example:
    class C(object):
        def __init__(self, value):
            self.value = value

    d = keydefaultdict(C)
    d[key] returns C(key)
    """
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            # pylint: disable=not-callable
            ret = self[key] = getattr(self, 'default_factory')(key)
            return ret


class CommonInfo(object):

    def __init__(self, id):
        self.id = id
        self.service_infos = dict()

    def add_service_info(self, service_info):
        if not service_info:
            return

        if not all(item in self.service_infos.items()
                   for item in service_info.items()):
            self.service_infos.update(service_info)


class PortInfo(CommonInfo):

    def __init__(self, port_id):
        super(PortInfo, self).__init__(port_id)

        self.ip_address = None
        self.mac_address = None
        self.network = None
        self.local_port = dict()

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.id == other.id)

    def __hash__(self):
        return hash(self.id)

    def set_local_port(self, linuxif):
        self.local_port = dict(linuxif=linuxif)

    def set_ip_mac_infos(self, ip_address, mac_address):
        self.ip_address = ip_address
        self.mac_address = mac_address

    def set_network(self, network):
        if not self.network:
            self.network = network
        else:
            LOG.warning('Network reference has already been set for port')


class NetworkInfo(CommonInfo):

    def __init__(self, network_id):
        super(NetworkInfo, self).__init__(network_id)

        self.gateway_info = b_const.NO_GW_INFO
        self.ports = set()

    def set_gateway_info(self, gateway_info):
        self.gateway_info = gateway_info


class BaseInfoManager(object):

    def __init__(self):
        # Store all ports level network and service informations
        self.ports_info = keydefaultdict(PortInfo)
        # Store all networks level network and service informations
        self.networks_info = keydefaultdict(NetworkInfo)

    def _get_network_port_infos(self, net_id, port_id):
        net_info = self.networks_info[net_id]
        port_info = self.ports_info[port_id]
        net_info.ports.add(port_info)
        port_info.set_network(net_info)

        return net_info, port_info

    def _remove_network_port_infos(self, net_id, port_id):
        port_info = self.ports_info.get(port_id)
        net_info = self.networks_info.get(net_id)

        if port_info:
            del self.ports_info[port_id]

        if net_info:
            net_info.ports.discard(port_info)
            if not net_info.ports:
                del self.networks_info[net_id]



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from oslo_utils import uuidutils

from networking_bagpipe.agent import bagpipe_bgp_agent

from networking_bagpipe.agent.bgpvpn import constants as bgpvpn_const

from neutron.plugins.ml2.drivers.linuxbridge.agent.common \
    import constants as lnx_agt_constants
from neutron.plugins.ml2.drivers.linuxbridge.agent \
    import linuxbridge_neutron_agent as lnx_agt
from neutron.plugins.ml2.drivers.openvswitch.agent.common \
    import constants as ovs_agt_constants
from neutron.plugins.ml2.drivers.openvswitch.agent \
    import ovs_agent_extension_api as ovs_ext_agt
from neutron.plugins.ml2.drivers.openvswitch.agent import vlanmanager

from neutron.tests import base
from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent import (
    ovs_test_base)

from neutron_lib import constants as n_const

PATCH_TUN_TO_MPLS_OFPORT = 1
PATCH_TUN_OFPORTS = [PATCH_TUN_TO_MPLS_OFPORT]

PATCH_INT_TO_MPLS_OFPORT = 5
PATCH_INT_OFPORTS = [PATCH_INT_TO_MPLS_OFPORT]

PATCH_MPLS_TO_TUN_OFPORT = 2
PATCH_MPLS_TO_INT_OFPORT = 6
PATCH_MPLS_OFPORTS = [PATCH_MPLS_TO_TUN_OFPORT, PATCH_MPLS_TO_INT_OFPORT]

PORT10_ID = uuidutils.generate_uuid()
PORT10 = {'id': PORT10_ID,
          'mac_address': '00:00:de:ad:be:ef',
          'ip_address': '10.0.0.2'}

PORT11 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:f0:0d',
          'ip_address': '10.0.0.3'}

NETWORK1 = {'id': uuidutils.generate_uuid(),
            'gateway_ip': '10.0.0.1'}

PORT20 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:be:ef',
          'ip_address': '20.0.0.2'}

PORT21 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:f0:0d',
          'ip_address': '20.0.0.3'}

NETWORK2 = {'id': uuidutils.generate_uuid(),
            'gateway_ip': '20.0.0.1'}

LOCAL_VLAN_MAP = {
    NETWORK1['id']: 31,
    NETWORK2['id']: 52
}

BAGPIPE_L2_RT1 = {'import_rt': 'BAGPIPE_L2:1',
                  'export_rt': 'BAGPIPE_L2:1'}

BAGPIPE_L2_RT2 = {'import_rt': 'BAGPIPE_L2:2',
                  'export_rt': 'BAGPIPE_L2:2'}

BGPVPN_L2_RT10 = {'import_rt': ['BGPVPN_L2:10'],
                  'export_rt': ['BGPVPN_L2:10']}

BGPVPN_L2_RT20 = {'import_rt': ['BGPVPN_L2:20'],
                  'export_rt': ['BGPVPN_L2:20']}

BGPVPN_L3_RT100 = {'import_rt': ['BGPVPN_L3:100'],
                   'export_rt': ['BGPVPN_L3:100']}

BGPVPN_L3_RT200 = {'import_rt': ['BGPVPN_L3:200'],
                   'export_rt': ['BGPVPN_L3:200']}


class DummyPort(object):
    def __init__(self, network, port, bgpvpn_port=False,
                 evpn=None, ipvpn=None):
        self.id = port['id']
        self.network_id = network['id']
        self.mac_address = port['mac_address']
        self.ip_address = port['ip_address']
        self.gateway_ip = network['gateway_ip']

        if bgpvpn_port:
            if evpn:
                self.l2vpn = copy.deepcopy(evpn)

            if ipvpn:
                self.l3vpn = copy.deepcopy(ipvpn)
        else:
            if evpn:
                self.evpn = copy.deepcopy(evpn)

            if ipvpn:
                self.ipvpn = copy.deepcopy(ipvpn)


class DummyVif(object):
    def __init__(self, ofport, port_name):
        self.ofport = ofport
        self.port_name = port_name


class DummyBGPVPN(object):
    def __init__(self, network, l2vpn=None, l3vpn=None, gateway_mac=None):
        self.id = uuidutils.generate_uuid()
        self.network_id = network['id']

        if l2vpn:
            self.l2vpn = copy.deepcopy(l2vpn)

        if l3vpn:
            self.l3vpn = copy.deepcopy(l3vpn)

        if gateway_mac:
            self.gateway_mac = gateway_mac


class RTList(list):

    def __eq__(self, other):
        return set(self) == set(other)


class BaseTestAgentExtension(object):

    agent_extension_class = None

    DUMMY_VIF10 = None
    DUMMY_VIF11 = None
    DUMMY_VIF20 = None
    DUMMY_VIF21 = None

    def _get_expected_local_port(self, vpn_type, network_id, port_id):
        raise NotImplementedError

    def _check_network_info(self, network_id, expected_size,
                            vpn_type=None, vpn_rts=None):
        if expected_size == 0:
            self.assertNotIn(network_id, self.agent_ext.networks_info,
                             "Network %s expected to have no ports left"
                             % network_id)
        else:
            self.assertIn(network_id, self.agent_ext.networks_info)
            network_info = self.agent_ext.networks_info[network_id]
            self.assertEqual(len(network_info.ports), expected_size,
                             "Network ports size not as expected")

            if vpn_type and vpn_rts:
                service_infos = network_info.service_infos
                self.assertIn(vpn_type, service_infos,
                              "No %s details found for network %s" %
                              (vpn_type, network_id))

                self.assertEqual(vpn_rts, service_infos[vpn_type],
                                 "%s details not matching %s for network %s" %
                                 (vpn_type, vpn_rts, network_id))


class BaseTestLinuxBridgeAgentExtension(base.BaseTestCase,
                                        BaseTestAgentExtension):

    def setUp(self):
        super(BaseTestLinuxBridgeAgentExtension, self).setUp()

        self.mocked_bagpipe_agent = mock.Mock(
            spec=bagpipe_bgp_agent.BaGPipeBGPAgent
        )
        self.mocked_bagpipe_agent.agent_type = n_const.AGENT_TYPE_LINUXBRIDGE
        self.mocked_bagpipe_agent.do_port_plug = mock.Mock()
        self.mocked_bagpipe_agent.do_port_plug_refresh = mock.Mock()

        self.agent_ext = self.agent_extension_class()
        self.connection = mock.Mock()

        agent_extension_api = mock.Mock()

        self.agent_ext.consume_api(agent_extension_api)
        with mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent.'
                        'BaGPipeBGPAgent.get_instance',
                        return_value=self.mocked_bagpipe_agent):
            self.agent_ext.initialize(self.connection,
                                      lnx_agt_constants.EXTENSION_DRIVER_TYPE,
                                      )

    def _get_expected_local_port(self, vpn_type, network_id, port_id):
        linuxbr = lnx_agt.LinuxBridgeManager.get_bridge_name(network_id)

        local_port = dict(
            linuxif=lnx_agt.LinuxBridgeManager.get_tap_device_name(port_id),
            evpnif=dict(linuxbr=linuxbr)
        )

        local_port.update(dict(
            ipvpnif=dict(
                local_port=dict(
                    linuxif=(linuxbr if vpn_type == bgpvpn_const.BGPVPN_L3
                             else None)
                )
            )
        ))

        return local_port


class BaseTestOVSAgentExtension(ovs_test_base.OVSOFCtlTestBase,
                                BaseTestAgentExtension):

    DUMMY_VIF10 = DummyVif(10, 'VIF10')
    DUMMY_VIF11 = DummyVif(11, 'VIF11')
    DUMMY_VIF20 = DummyVif(20, 'VIF20')
    DUMMY_VIF21 = DummyVif(21, 'VIF21')

    def setUp(self):
        super(BaseTestOVSAgentExtension, self).setUp()

        self.mocked_bagpipe_agent = mock.Mock(
            spec=bagpipe_bgp_agent.BaGPipeBGPAgent
        )
        self.mocked_bagpipe_agent.agent_type = n_const.AGENT_TYPE_OVS
        self.mocked_bagpipe_agent.do_port_plug = mock.Mock()
        self.mocked_bagpipe_agent.do_port_plug_refresh = mock.Mock()

        self.agent_ext = self.agent_extension_class()
        self.connection = mock.Mock()

        int_br = self.br_int_cls("br-int")
        int_br.add_patch_port = mock.Mock(side_effect=PATCH_INT_OFPORTS)
        int_br.get_port_ofport = mock.Mock()
        int_br.add_flow = mock.Mock()
        int_br.delete_flows = mock.Mock()

        tun_br = self.br_tun_cls("br-tun")
        tun_br.add_patch_port = mock.Mock(side_effect=PATCH_TUN_OFPORTS)
        tun_br.get_port_ofport = mock.Mock()
        tun_br.add_flow = mock.Mock()
        tun_br.delete_flows = mock.Mock()

        agent_extension_api = ovs_ext_agt.OVSAgentExtensionAPI(int_br,
                                                               tun_br)

        self.agent_ext.consume_api(agent_extension_api)

        with mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                        'bridge_exists', return_value=True), \
                mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                           'add_patch_port', side_effect=PATCH_MPLS_OFPORTS), \
                mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent.'
                           'BaGPipeBGPAgent.get_instance',
                           return_value=self.mocked_bagpipe_agent):
            self.agent_ext.initialize(self.connection,
                                      ovs_agt_constants.EXTENSION_DRIVER_TYPE,
                                      )

        self.vlan_manager = vlanmanager.LocalVlanManager()
        for net_id, vlan in LOCAL_VLAN_MAP.items():
            try:
                self.vlan_manager.add(net_id, vlan, None, None, None)
            except vlanmanager.MappingAlreadyExists:
                pass

    def _get_expected_local_port(self, vpn_type, network_id, port_id):
        vlan = self.vlan_manager.get(network_id).vlan

        local_port = dict(
            linuxif='%s:%s' % (bgpvpn_const.LINUXIF_PREFIX, vlan),
            evpnif=None,
            ipvpnif=dict(
                local_port=dict(
                    ovs=dict(plugged=True,
                             port_number=PATCH_MPLS_TO_TUN_OFPORT,
                             vlan=vlan)
                )
            )
        )

        return local_port



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock

from networking_bagpipe.agent.common import constants as b_const

from networking_bagpipe.agent import bagpipe_linuxbridge_neutron_agent as\
    linuxbridge_agent

from networking_bagpipe.tests.unit.agent import base


class TestBaGPipeAgentExtensionMixin(object):

    def _format_rt_as_list(self, rt):
        return {k: [v] for k, v in rt.items()}

    def _get_vpn_info(self, vpn_type, vpn_if, rts):
        vpn_info = {vpn_type: vpn_if}
        vpn_info[vpn_type].update(self._format_rt_as_list(rts))

        return vpn_info

    def test_bagpipe_l2_attach_single_port(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__

        self.agent_ext.bagpipe_port_attach(None, copy.copy(dummy_port10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        self._check_network_info(base.NETWORK1['id'],
                                 1,
                                 b_const.EVPN,
                                 base.BAGPIPE_L2_RT1)

        # Verify build callback attachments
        local_port = self._get_expected_local_port(b_const.EVPN,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])

        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **self._get_vpn_info(b_const.EVPN,
                                     local_port['evpnif'],
                                     base.BAGPIPE_L2_RT1)
            ),
            self.agent_ext.build_bagpipe_l2_attach_info(base.PORT10['id'])
        )

    def test_bagpipe_l2_attach_multiple_ports_same_network(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__

        self.agent_ext.bagpipe_port_attach(None, copy.copy(dummy_port10))
        self.agent_ext.bagpipe_port_attach(None, copy.copy(dummy_port11))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])]
        )

        self._check_network_info(base.NETWORK1['id'],
                                 2,
                                 b_const.EVPN,
                                 base.BAGPIPE_L2_RT1)

        # Verify build callback attachments
        for port in [base.PORT10, base.PORT11]:
            local_port = self._get_expected_local_port(b_const.EVPN,
                                                       base.NETWORK1['id'],
                                                       port['id'])

            self.assertEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ip_address=port['ip_address'],
                    mac_address=port['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.EVPN,
                                         local_port['evpnif'],
                                         base.BAGPIPE_L2_RT1)
                ),
                self.agent_ext.build_bagpipe_l2_attach_info(port['id'])
            )

    def test_bagpipe_l2_attach_multiple_ports_different_networks(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                      evpn=base.BAGPIPE_L2_RT2).__dict__

        self.agent_ext.bagpipe_port_attach(None, copy.copy(dummy_port10))
        self.agent_ext.bagpipe_port_attach(None, copy.copy(dummy_port20))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT20['id'])]
        )

        for network_id, evpn_rt in [
                (base.NETWORK1['id'], base.BAGPIPE_L2_RT1),
                (base.NETWORK2['id'], base.BAGPIPE_L2_RT2)]:
            self._check_network_info(network_id,
                                     1,
                                     b_const.EVPN,
                                     evpn_rt)

        # Verify build callback attachments
        for network, port, rts in [(base.NETWORK1, base.PORT10,
                                    base.BAGPIPE_L2_RT1),
                                   (base.NETWORK2, base.PORT20,
                                    base.BAGPIPE_L2_RT2)]:
            local_port = self._get_expected_local_port(b_const.EVPN,
                                                       network['id'],
                                                       port['id'])

            self.assertEqual(
                dict(
                    network_id=network['id'],
                    ip_address=port['ip_address'],
                    mac_address=port['mac_address'],
                    gateway_ip=network['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.EVPN,
                                         local_port['evpnif'],
                                         rts)
                ),
                self.agent_ext.build_bagpipe_l2_attach_info(port['id'])
            )

    def test_bagpipe_l2_detach_single_port(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        self.agent_ext.bagpipe_port_attach(None, dummy_port10)

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.bagpipe_port_detach(None, dummy_detach10)

        local_port = self._get_expected_local_port(b_const.EVPN,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])
        detach_info = {
            b_const.EVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info)]
        )

        self._check_network_info(base.NETWORK1['id'], 0)
        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)

    def test_bagpipe_l2_detach_multiple_ports_same_network(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach11 = dict(id=base.PORT11['id'],
                              network_id=base.NETWORK1['id'])

        # Attach 2 ports on network 1
        self.agent_ext.bagpipe_port_attach(None, dummy_port10)
        self.agent_ext.bagpipe_port_attach(None, dummy_port11)

        self.mocked_bagpipe_agent.reset_mock()

        # Detach 1 port from network 1
        self.agent_ext.bagpipe_port_detach(None, dummy_detach10)

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'],
                                 1,
                                 b_const.EVPN,
                                 base.BAGPIPE_L2_RT1)

        # Detach remaining port from network 1
        self.agent_ext.bagpipe_port_detach(None, dummy_detach11)

        local_port10 = self._get_expected_local_port(b_const.EVPN,
                                                     base.NETWORK1['id'],
                                                     base.PORT10['id'])
        detach_info10 = {
            b_const.EVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port10['linuxif'])
            }
        }

        local_port11 = self._get_expected_local_port(b_const.EVPN,
                                                     base.NETWORK1['id'],
                                                     base.PORT11['id'])
        detach_info11 = {
            b_const.EVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT11['ip_address'],
                'mac_address': base.PORT11['mac_address'],
                'local_port': dict(linuxif=local_port11['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info10),
             mock.call(base.PORT11['id'], detach_info11)]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 0)
        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)

    def test_bagpipe_l2_detach_multiple_ports_different_networks(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach11 = dict(id=base.PORT11['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                      evpn=base.BAGPIPE_L2_RT2).__dict__
        dummy_detach20 = dict(id=base.PORT20['id'],
                              network_id=base.NETWORK2['id'])

        dummy_port21 = base.DummyPort(base.NETWORK2, base.PORT21,
                                      evpn=base.BAGPIPE_L2_RT2).__dict__
        dummy_detach21 = dict(id=base.PORT21['id'],
                              network_id=base.NETWORK2['id'])

        # Attach 2 ports on network 1
        self.agent_ext.bagpipe_port_attach(None, dummy_port10)
        self.agent_ext.bagpipe_port_attach(None, dummy_port11)

        # Attach 2 ports on network 2
        self.agent_ext.bagpipe_port_attach(None, dummy_port20)
        self.agent_ext.bagpipe_port_attach(None, dummy_port21)

        self.mocked_bagpipe_agent.reset_mock()

        # Detach 1 port from each network
        self.agent_ext.bagpipe_port_detach(None, dummy_detach10)
        self.agent_ext.bagpipe_port_detach(None, dummy_detach20)

        local_port10 = self._get_expected_local_port(b_const.EVPN,
                                                     base.NETWORK1['id'],
                                                     base.PORT10['id'])
        detach_info10 = {
            b_const.EVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port10['linuxif'])
            }
        }

        local_port20 = self._get_expected_local_port(b_const.EVPN,
                                                     base.NETWORK2['id'],
                                                     base.PORT20['id'])
        detach_info20 = {
            b_const.EVPN: {
                'network_id': base.NETWORK2['id'],
                'ip_address': base.PORT20['ip_address'],
                'mac_address': base.PORT20['mac_address'],
                'local_port': dict(linuxif=local_port20['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info10),
             mock.call(base.PORT20['id'], detach_info20)]
        )

        # Verify attachments list consistency
        for network_id, evpn_rt in [
                (base.NETWORK1['id'], base.BAGPIPE_L2_RT1),
                (base.NETWORK2['id'], base.BAGPIPE_L2_RT2)]:
            self._check_network_info(network_id,
                                     1,
                                     b_const.EVPN,
                                     evpn_rt)

        self.mocked_bagpipe_agent.reset_mock()

        # Detach remaining port from each network
        self.agent_ext.bagpipe_port_detach(None, dummy_detach11)
        self.agent_ext.bagpipe_port_detach(None, dummy_detach21)

        local_port11 = self._get_expected_local_port(b_const.EVPN,
                                                     base.NETWORK1['id'],
                                                     base.PORT11['id'])
        detach_info11 = {
            b_const.EVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT11['ip_address'],
                'mac_address': base.PORT11['mac_address'],
                'local_port': dict(linuxif=local_port11['linuxif'])
            }
        }

        local_port21 = self._get_expected_local_port(b_const.EVPN,
                                                     base.NETWORK2['id'],
                                                     base.PORT21['id'])
        detach_info21 = {
            b_const.EVPN: {
                'network_id': base.NETWORK2['id'],
                'ip_address': base.PORT21['ip_address'],
                'mac_address': base.PORT21['mac_address'],
                'local_port': dict(linuxif=local_port21['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT11['id'], detach_info11),
             mock.call(base.PORT21['id'], detach_info21)]
        )

        # Verify attachments list consistency
        for network_id in [base.NETWORK1['id'], base.NETWORK2['id']]:
            self._check_network_info(network_id, 0)

        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)


class LinuxBridgeAgentExtensionTest(base.BaseTestLinuxBridgeAgentExtension,
                                    TestBaGPipeAgentExtensionMixin):

    agent_extension_class = linuxbridge_agent.BagpipeAgentExtension



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple

# bagpipe-bgp VPN types
EVPN = 'evpn'
IPVPN = 'ipvpn'
VPN_TYPES = [EVPN, IPVPN]

RT_IMPORT = 'import_rt'
RT_EXPORT = 'export_rt'
RT_TYPES = [RT_IMPORT, RT_EXPORT]

GatewayInfo = namedtuple('GatewayInfo', ['mac', 'ip'])
NO_GW_INFO = GatewayInfo(None, None)



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from networking_bagpipe.agent import bagpipe_bgp_agent as agent
from networking_bagpipe.agent.common import constants as b_const

from networking_bagpipe.tests.unit.agent.common \
    import constants as const

from neutron.tests import base


class TestBaGPipeBGPAgentSingleService(base.BaseTestCase):

    def setUp(self):
        super(TestBaGPipeBGPAgentSingleService, self).setUp()

        self.agent = agent.BaGPipeBGPAgent('SINGLE_SERVICE')
        self.agent._send_attach_local_port = mock.Mock()
        self.agent._send_detach_local_port = mock.Mock()

        self.service1 = mock.Mock(name='SERVICE1')
        self.agent.register_build_callback(self.service1.name,
                                           self.service1._build_port_info)

    def _test_port_plug(self, vpn_type, rts):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update({vpn_type: rts})
        port_info1.update(**const.NETWORK_INFO1)

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           vpn_type),
                vpn_type=vpn_type,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts
            ))
        ])

    def test_evpn_port_plug(self):
        self._test_port_plug(b_const.EVPN, const.EVPN_RT1)

    def test_ipvpn_port_plug(self):
        self._test_port_plug(b_const.IPVPN, const.IPVPN_RT100)

    def test_evpn2ipvpn_port_plug(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(evpn=const.EVPN_RT1,
                               ipvpn=const.IPVPN_RT100,
                               **const.NETWORK_INFO1))

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **const.EVPN_RT1
            )),
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.IPVPN),
                vpn_type=b_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port={
                    b_const.EVPN: {
                        'id': '%s_%s' % (const.NETWORK_INFO1['network_id'],
                                         b_const.EVPN)
                    }
                },
                **const.IPVPN_RT100
            ))
        ])

    def test_epvpn_port_plug_updated_local_port(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(evpn=dict(local_port=const.UPDATED_LOCAL_PORT1,
                                         **const.EVPN_RT1),
                               **const.NETWORK_INFO1))

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.UPDATED_LOCAL_PORT1,
                **const.EVPN_RT1
            ))
        ])

    def test_epvpn_port_plug_gateway_mac_added(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(evpn=dict(gateway_mac=const.GW_MAC_PORT1,
                                         **const.EVPN_RT1),
                               **const.NETWORK_INFO1))

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                gateway_mac=const.GW_MAC_PORT1,
                **const.EVPN_RT1
            ))
        ])

    def test_epvpn_port_plug_static_routes(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(evpn=dict(static_routes=[const.STATIC_ROUTE1],
                                         **const.EVPN_RT1),
                               **const.NETWORK_INFO1))

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.STATIC_ROUTE1,
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                advertise_subnet=True,
                **const.EVPN_RT1
            )),
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **const.EVPN_RT1
            )),
        ])

    def _test_port_plug_refresh_without_detach(self, vpn_type, rts):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update({vpn_type: rts})
        port_info1.update(**const.NETWORK_INFO1)

        detach_info1 = {
            vpn_type: dict(network_id=const.NETWORK_INFO1['network_id'],
                           **const.PORT_INFO1)
        }

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           vpn_type),
                vpn_type=vpn_type,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **rts
            ))
        ])

        self.agent._send_detach_local_port.assert_not_called()

    def test_evpn_port_plug_refresh_without_detach(self):
        self._test_port_plug_refresh_without_detach(b_const.EVPN,
                                                    const.EVPN_RT1)

    def test_ipvpn_port_plug_refresh_without_detach(self):
        self._test_port_plug_refresh_without_detach(b_const.IPVPN,
                                                    const.IPVPN_RT100)

    def _test_port_plug_refresh_with_detach(self, vpn_type, rts):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update({vpn_type: rts})
        port_info1.update(**const.NETWORK_INFO1)

        detach_info1 = {
            vpn_type: dict(network_id=const.NETWORK_INFO1['network_id'],
                           **const.PORT_INFO1)
        }

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_not_called()

        self.agent._send_detach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           vpn_type),
                vpn_type=vpn_type,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                local_port=const.PORT_INFO1['local_port']
            ))
        ])

    def test_evpn_port_plug_refresh_with_detach(self):
        self._test_port_plug_refresh_without_detach(b_const.EVPN,
                                                    const.EVPN_RT1)

    def test_ipvpn_port_plug_refresh_with_detach(self):
        self._test_port_plug_refresh_without_detach(b_const.IPVPN,
                                                    const.IPVPN_RT100)

    def test_evpn2ipvpn_port_plug_refresh_with_detach(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(evpn=const.EVPN_RT1,
                               **const.NETWORK_INFO1))

        detach_info1 = {
            b_const.IPVPN: dict(network_id=const.NETWORK_INFO1['network_id'],
                                **const.PORT_INFO1)
        }

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **const.EVPN_RT1
            ))
        ])

        self.agent._send_detach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.IPVPN),
                vpn_type=b_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                local_port={
                    b_const.EVPN: {
                        'id': '%s_%s' % (const.NETWORK_INFO1['network_id'],
                                         b_const.EVPN)
                    }
                }
            ))
        ])


class RTList(list):

    def __eq__(self, other):
        return set(self) == set(other)


class TestBaGPipeBGPAgentMultipleServices(base.BaseTestCase):

    def setUp(self):
        super(TestBaGPipeBGPAgentMultipleServices, self).setUp()

        self.agent = agent.BaGPipeBGPAgent('MULTIPLE_SERVICES')
        self.agent._send_attach_local_port = mock.Mock()
        self.agent._send_detach_local_port = mock.Mock()

        self.service1 = mock.Mock(name='SERVICE1')
        self.agent.register_build_callback(self.service1.name,
                                           self.service1._build_port_info)

        self.service2 = mock.Mock(name='SERVICE2')
        self.agent.register_build_callback(self.service2.name,
                                           self.service2._build_port_info)

    def _merge_rts(self, rt1, rt2):
        return {k: RTList(rt1[k] + rt2[k]) for k in rt1}

    def test_evpns_port_plug(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(evpn=const.EVPN_RT1,
                               **const.NETWORK_INFO1))

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        port_info1bis = copy.copy(const.PORT_INFO1)
        port_info1bis.update(dict(evpn=const.EVPN_RT2,
                                  **const.NETWORK_INFO1))

        self.service2._build_port_info.return_value = (
            copy.deepcopy(port_info1bis)
        )

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **self._merge_rts(const.EVPN_RT1, const.EVPN_RT2)
            ))
        ])

    def test_ipvpns_port_plug(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(ipvpn=const.IPVPN_RT100,
                               **const.NETWORK_INFO1))

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        port_info1bis = copy.copy(const.PORT_INFO1)
        port_info1bis.update(dict(ipvpn=const.IPVPN_RT200,
                                  **const.NETWORK_INFO1))

        self.service2._build_port_info.return_value = (
            copy.deepcopy(port_info1bis)
        )

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.IPVPN),
                vpn_type=b_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **self._merge_rts(const.IPVPN_RT100, const.IPVPN_RT200)
            ))
        ])

    def test_evpn2ipvpn_port_plug(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(evpn=const.EVPN_RT1,
                               **const.NETWORK_INFO1))

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        port_info1bis = copy.copy(const.PORT_INFO1)
        port_info1bis.update(dict(ipvpn=const.IPVPN_RT100,
                                  **const.NETWORK_INFO1))

        self.service2._build_port_info.return_value = (
            copy.deepcopy(port_info1bis)
        )

        self.agent.do_port_plug(None)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **const.EVPN_RT1
            )),
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.IPVPN),
                vpn_type=b_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port={
                    b_const.EVPN: {
                        'id': '%s_%s' % (const.NETWORK_INFO1['network_id'],
                                         b_const.EVPN)
                    }
                },
                **const.IPVPN_RT100
            ))
        ])

    def test_evpns_port_plug_refresh_without_detach(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(evpn=const.EVPN_RT1,
                               **const.NETWORK_INFO1))

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.service2._build_port_info.return_value = {}

        detach_info1 = {
            b_const.EVPN: dict(network_id=const.NETWORK_INFO1['network_id'],
                               **const.PORT_INFO1)
        }

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **const.EVPN_RT1
            ))
        ])

    def test_ipvpns_port_plug_refresh_without_detach(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(ipvpn=const.IPVPN_RT100,
                               **const.NETWORK_INFO1))

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.service2._build_port_info.return_value = {}

        detach_info1 = {
            b_const.IPVPN: dict(network_id=const.NETWORK_INFO1['network_id'],
                                **const.PORT_INFO1)
        }

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.IPVPN),
                vpn_type=b_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **const.IPVPN_RT100
            ))
        ])

    def test_evpn2ipvpn_port_plug_refresh_with_detach(self):
        port_info1 = copy.copy(const.PORT_INFO1)
        port_info1.update(dict(evpn=const.EVPN_RT1,
                               **const.NETWORK_INFO1))

        detach_info1 = {
            b_const.IPVPN: dict(network_id=const.NETWORK_INFO1['network_id'],
                                **const.PORT_INFO1)
        }

        self.service1._build_port_info.return_value = (
            copy.deepcopy(port_info1)
        )

        self.service2._build_port_info.return_value = {}

        self.agent.do_port_plug_refresh(None, detach_info1)

        self.agent._send_attach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.EVPN),
                vpn_type=b_const.EVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                gateway_ip=const.NETWORK_INFO1['gateway_ip'],
                local_port=const.PORT_INFO1['local_port'],
                **const.EVPN_RT1
            ))
        ])

        self.agent._send_detach_local_port.assert_has_calls([
            mock.call(dict(
                vpn_instance_id='%s_%s' % (const.NETWORK_INFO1['network_id'],
                                           b_const.IPVPN),
                vpn_type=b_const.IPVPN,
                ip_address=const.PORT_INFO1['ip_address'],
                mac_address=const.PORT_INFO1['mac_address'],
                local_port={
                    b_const.EVPN: {
                        'id': '%s_%s' % (const.NETWORK_INFO1['network_id'],
                                         b_const.EVPN)
                    }
                }
            ))
        ])



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from networking_bagpipe.bagpipe_bgp import constants as consts

from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp.vpn import manager
from networking_bagpipe.tests.unit.bagpipe_bgp import base as t


REDIRECTED_INSTANCE_ID1 = 'redirected-id1'
REDIRECTED_INSTANCE_ID2 = 'redirected-id2'

MAC = "00:00:de:ad:be:ef"
IP = "10.0.0.1/32"
BRIDGE_NAME = "br-test"
LOCAL_PORT = {'linuxif': 'tap1'}
VPN_EXT_ID = 1
GW_IP = "10.0.0.1"
GW_MASK = 24
VNID = 255


class TestVPNManager(t.TestCase):

    def setUp(self):
        super(TestVPNManager, self).setUp()
        mock.patch("networking_bagpipe.bagpipe_bgp.vpn.dataplane_drivers."
                   "instantiate_dataplane_drivers",
                   return_value={
                       'evpn': mock.Mock(),
                       'ipvpn': mock.Mock()
                   }).start()
        self.manager = manager.VPNManager()

    def tearDown(self):
        super(TestVPNManager, self).tearDown()
        self.manager.stop()

    def test_redirect_traffic_single_instance(self):
        redirect_instance = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID1, consts.IPVPN, t._rt_to_string(t.RT5)
        )

        # Check some VPN manager and redirect instance lists consistency
        self.assertIn(
            manager.redirect_instance_extid(consts.IPVPN,
                                            t._rt_to_string(t.RT5)),
            self.manager.vpn_instances)
        self.assertIn(REDIRECTED_INSTANCE_ID1,
                      redirect_instance.redirected_instances)

    def test_redirect_traffic_multiple_instance(self):
        redirect_instance_1 = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID1, consts.IPVPN, t._rt_to_string(t.RT5)
        )
        redirect_instance_2 = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID2, consts.IPVPN, t._rt_to_string(t.RT5)
        )

        # Check that same redirect instance is returned
        self.assertEqual(redirect_instance_2, redirect_instance_1)
        # Check some VPN manager and redirect instance lists consistency
        self.assertIn(
            manager.redirect_instance_extid(consts.IPVPN,
                                            t._rt_to_string(t.RT5)),
            self.manager.vpn_instances)
        self.assertIn(REDIRECTED_INSTANCE_ID1,
                      redirect_instance_1.redirected_instances)
        self.assertIn(REDIRECTED_INSTANCE_ID2,
                      redirect_instance_1.redirected_instances)

    def test_stop_redirect_traffic_multiple_instance(self):
        redirect_instance = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID1, consts.IPVPN, t._rt_to_string(t.RT5)
        )
        self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID2, consts.IPVPN, t._rt_to_string(t.RT5)
        )

        # Check some VPN manager and redirect instance lists consistency
        self.manager.stop_redirect_to_vpn(REDIRECTED_INSTANCE_ID2,
                                          consts.IPVPN, t._rt_to_string(t.RT5))

        self.assertNotIn(REDIRECTED_INSTANCE_ID2,
                         redirect_instance.redirected_instances)

        self.manager.stop_redirect_to_vpn(REDIRECTED_INSTANCE_ID1,
                                          consts.IPVPN, t._rt_to_string(t.RT5))

        self.assertTrue(not self.manager.vpn_instances)

    def test_plug_vif_to_vpn_with_forced_vni(self):
        with mock.patch.object(self.manager, "_get_vpn_instance") as mock_get_vpn_instance, \
                mock.patch.object(utils, "convert_route_targets"):
            self.manager.plug_vif_to_vpn(vpn_instance_id=VPN_EXT_ID,
                                         vpn_type=consts.EVPN,
                                         import_rt=[t.RT1],
                                         export_rt=[t.RT1],
                                         mac_address=MAC,
                                         ip_address=IP,
                                         gateway_ip=GW_IP,
                                         local_port=LOCAL_PORT,
                                         linuxbr=BRIDGE_NAME,
                                         vni=VNID)
        mock_get_vpn_instance.assert_called_once_with(
            VPN_EXT_ID, consts.EVPN, mock.ANY, mock.ANY, GW_IP, mock.ANY,
            None, None, None, linuxbr=BRIDGE_NAME, vni=VNID)

    def test_plug_vif_to_vpn_without_forced_vni(self):
        with mock.patch.object(self.manager, "_get_vpn_instance") as mock_get_vpn_instance, \
                mock.patch.object(utils, "convert_route_targets"):
            self.manager.plug_vif_to_vpn(vpn_instance_id=VPN_EXT_ID,
                                         vpn_type=consts.EVPN,
                                         import_rt=[t.RT1],
                                         export_rt=[t.RT1],
                                         mac_address=MAC,
                                         ip_address=IP,
                                         gateway_ip=GW_IP,
                                         local_port=LOCAL_PORT,
                                         linuxbr=BRIDGE_NAME)

        mock_get_vpn_instance.assert_called_once_with(
            VPN_EXT_ID, consts.EVPN, mock.ANY, mock.ANY, GW_IP, mock.ANY,
            None, None, None, linuxbr=BRIDGE_NAME)

    def test_get_vpn_instance_with_forced_vni(self):
        vpn_instance = self.manager._get_vpn_instance(VPN_EXT_ID,
                                                      consts.IPVPN,
                                                      [], [],
                                                      GW_IP, GW_MASK,
                                                      None, None,
                                                      vni=VNID)

        self.assertEqual(VNID, vpn_instance.instance_label,
                         "VPN instance label should be forced to VNID")

    def test_get_vpn_instance_without_forced_vni(self):
        vpn_instance = self.manager._get_vpn_instance(VPN_EXT_ID,
                                                      consts.IPVPN,
                                                      [], [],
                                                      GW_IP, GW_MASK,
                                                      None, None)

        self.assertIsNot(0, vpn_instance.instance_label,
                         "VPN instance label should be assigned locally")

    @mock.patch('networking_bagpipe.bagpipe_bgp.engine.bgp_manager.Manager')
    def test_manager_stop(self, mocked_bgp_manager):
        self.manager._get_vpn_instance("TEST_VPN_INSTANCE", consts.IPVPN,
                                       [t.RT1], [t.RT1], "192.168.0.1", 24,
                                       {}, {})
        self.manager.stop()
        self.assertTrue(not self.manager.vpn_instances)



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock

from networking_bagpipe.agent.common import constants as b_const

from networking_bagpipe.agent.bgpvpn import agent_extension as bagpipe_agt_ext
from networking_bagpipe.agent.bgpvpn import constants as bgpvpn_const

from networking_bagpipe.tests.unit.agent import base


def update_delete_rpc(d):
    # update dict d so that the content looks like an RPC generated by
    # bagpipe bgpvpn driver on neutron server-side
    d.setdefault('id', 'id-foo')
    d.setdefault('network_id', 'net-id-foo')
    d.setdefault('gateway_mac', None)
    return d


class TestBgpvpnAgentExtensionMixin(object):

    def _get_vpn_info(self, vpn_type, vpn_if, rts, fallback=None,
                      more_vpn_info=None):
        vpn_info = {vpn_type: vpn_if}
        vpn_info[vpn_type].update(rts)

        if fallback:
            vpn_info[vpn_type].update(dict(fallback=fallback))

        vpn_info[vpn_type].update(more_vpn_info or {})

        return copy.deepcopy(vpn_info)

    def test_update_bgpvpn_no_plugged_ports(self):
        dummy_bgpvpn1 = base.DummyBGPVPN(
            base.NETWORK1,
            l2vpn=base.BGPVPN_L2_RT10).__dict__

        update_delete_rpc(dummy_bgpvpn1)

        self.agent_ext.update_bgpvpn(None, dummy_bgpvpn1)

        self.assertEqual(0, self.mocked_bagpipe_agent.do_port_plug.call_count,
                         "Do port plug mustn't be called")

    def test_update_bgpvpn_already_plugged_ports(self):
        dummy_port10 = base.DummyPort(base.NETWORK1,
                                      base.PORT10).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1,
                                      base.PORT11).__dict__

        dummy_bgpvpn1 = base.DummyBGPVPN(
            base.NETWORK1,
            l3vpn=base.BGPVPN_L3_RT100).__dict__

        update_delete_rpc(dummy_bgpvpn1)

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))
        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port11))

        self.assertEqual(self.mocked_bagpipe_agent.do_port_plug.call_count, 0)

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'],
                                 2)

        self.agent_ext.update_bgpvpn(None, dummy_bgpvpn1)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'],
                                 2,
                                 bgpvpn_const.BGPVPN_L3,
                                 base.BGPVPN_L3_RT100)

        # Verify build callback attachments
        for port in [base.PORT10, base.PORT11]:
            local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                       base.NETWORK1['id'],
                                                       port['id'])

            self.assertEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ip_address=port['ip_address'],
                    mac_address=port['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.IPVPN,
                                         local_port['ipvpnif'],
                                         base.BGPVPN_L3_RT100)
                ),
                self.agent_ext.build_bgpvpn_attach_info(port['id'])
            )

    def test_delete_bgpvpn_after_update_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1,
                                      base.PORT10).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1,
                                      base.PORT11).__dict__

        dummy_bgpvpn1 = base.DummyBGPVPN(
            base.NETWORK1,
            l3vpn=base.BGPVPN_L3_RT100).__dict__

        update_delete_rpc(dummy_bgpvpn1)

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))
        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port11))

        self.assertEqual(self.mocked_bagpipe_agent.do_port_plug.call_count, 0)

        self.agent_ext.update_bgpvpn(None, copy.deepcopy(dummy_bgpvpn1))

        self.assertEqual(self.mocked_bagpipe_agent.do_port_plug.call_count, 2)
        self.mocked_bagpipe_agent.do_port_plug.reset_mock()

        self.agent_ext.delete_bgpvpn(None, copy.deepcopy(dummy_bgpvpn1))

        self.mocked_bagpipe_agent.do_port_plug.assert_not_called()
        self.assertEqual(
            self.mocked_bagpipe_agent.do_port_plug_refresh.call_count, 2)

    def test_update_bgpvpn_before_port_detach(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_bgpvpn1 = base.DummyBGPVPN(
            base.NETWORK1,
            l3vpn=base.BGPVPN_L3_RT100).__dict__

        update_delete_rpc(dummy_bgpvpn1)

        self.agent_ext.bgpvpn_port_attach(None, dummy_port10)

        self.agent_ext.update_bgpvpn(None, dummy_bgpvpn1)

        self.agent_ext.bgpvpn_port_detach(None, dummy_detach10)

        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])
        detach_info = {
            b_const.IPVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info)]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 0)

        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)

    def test_update_bgpvpn_same_vpn_types(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__

        ipvpn_rts = ({k: base.BGPVPN_L3_RT100[k] + base.BGPVPN_L3_RT200[k]
                      for k in b_const.RT_TYPES})

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l3vpn=ipvpn_rts).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))

        self.agent_ext.update_bgpvpn(None, dummy_bgpvpn1)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        self._check_network_info(base.NETWORK1['id'],
                                 1,
                                 bgpvpn_const.BGPVPN_L3,
                                 ipvpn_rts)

        # Verify build callback attachments
        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])
        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **self._get_vpn_info(b_const.IPVPN,
                                     local_port['ipvpnif'],
                                     ipvpn_rts)
            ),
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

    def test_delete_bgpvpn_remaining_plugged_ports(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l3vpn=base.BGPVPN_L3_RT100).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))
        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port11))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'],
                                 2,
                                 bgpvpn_const.BGPVPN_L3,
                                 base.BGPVPN_L3_RT100)

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.delete_bgpvpn(None, dummy_bgpvpn1)

        local_port10 = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                     base.NETWORK1['id'],
                                                     base.PORT10['id'])
        detach_info10 = {
            b_const.IPVPN: dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                local_port=dict(linuxif=local_port10['linuxif'])
            )
        }

        local_port11 = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                     base.NETWORK1['id'],
                                                     base.PORT11['id'])
        detach_info11 = {
            b_const.IPVPN: dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT11['ip_address'],
                mac_address=base.PORT11['mac_address'],
                local_port=dict(linuxif=local_port11['linuxif'])
            )
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info10),
             mock.call(base.PORT11['id'], detach_info11)],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 2)

        # Verify build callback attachments
        self.assertEqual(
            {},
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

    def test_delete_bgpvpn_remaining_plugged_ports_after_update(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__

        update_rts = ({k: base.BGPVPN_L3_RT100[k] + base.BGPVPN_L3_RT200[k]
                       for k in b_const.RT_TYPES})

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l3vpn=update_rts).__dict__

        dummy_bgpvpn1bis = base.DummyBGPVPN(
            base.NETWORK1,
            l3vpn=base.BGPVPN_L3_RT200).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))
        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port11))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'],
                                 2,
                                 bgpvpn_const.BGPVPN_L3,
                                 base.BGPVPN_L3_RT100)

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.update_bgpvpn(None, dummy_bgpvpn1)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])],
            any_order=True
        )

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.delete_bgpvpn(None, dummy_bgpvpn1bis)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'],
                                 2,
                                 bgpvpn_const.BGPVPN_L3,
                                 base.BGPVPN_L3_RT100)

        # Verify build callback attachments
        for port in [base.PORT10, base.PORT10]:
            local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                       base.NETWORK1['id'],
                                                       port['id'])

            self.assertEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ip_address=port['ip_address'],
                    mac_address=port['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.IPVPN,
                                         local_port['ipvpnif'],
                                         base.BGPVPN_L3_RT100)
                ),
                self.agent_ext.build_bgpvpn_attach_info(port['id'])
            )

    def test_delete_bgpvpn_remaining_plugged_ports_after_import_update(self):
        bgpvpn_l3_import = copy.copy(base.BGPVPN_L3_RT200)
        bgpvpn_l3_import['export_rt'] = []
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__

        update_rts = ({k: base.BGPVPN_L3_RT100[k] + bgpvpn_l3_import[k]
                       for k in b_const.RT_TYPES})

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l3vpn=update_rts).__dict__

        dummy_bgpvpn1bis = base.DummyBGPVPN(
            base.NETWORK1,
            l3vpn=base.BGPVPN_L3_RT100).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))
        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port11))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'],
                                 2,
                                 bgpvpn_const.BGPVPN_L3,
                                 base.BGPVPN_L3_RT100)

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.update_bgpvpn(None, dummy_bgpvpn1)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])],
            any_order=True
        )

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.delete_bgpvpn(None, dummy_bgpvpn1bis)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'],
                                 2,
                                 bgpvpn_const.BGPVPN_L3,
                                 bgpvpn_l3_import)

        # Verify build callback attachments
        for port in [base.PORT10, base.PORT10]:
            local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                       base.NETWORK1['id'],
                                                       port['id'])

            self.assertEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ip_address=port['ip_address'],
                    mac_address=port['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.IPVPN,
                                         local_port['ipvpnif'],
                                         bgpvpn_l3_import)
                ),
                self.agent_ext.build_bgpvpn_attach_info(port['id'])
            )

    def test_delete_bgpvpn_no_plugged_ports(self):
        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l3vpn=base.BGPVPN_L3_RT100).__dict__

        self.agent_ext.delete_bgpvpn(None, dummy_bgpvpn1)

        self.assertEqual(0, self.mocked_bagpipe_agent.do_port_plug.call_count,
                         "Do port plug mustn't be called")

    def test_delete_bgpvpn_had_plugged_ports(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach11 = dict(id=base.PORT11['id'],
                              network_id=base.NETWORK1['id'])

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l3vpn=base.BGPVPN_L3_RT100).__dict__

        self.agent_ext.bgpvpn_port_attach(None, dummy_port10)
        self.agent_ext.bgpvpn_port_attach(None, dummy_port11)

        self.agent_ext.bgpvpn_port_detach(None, dummy_detach10)
        self.agent_ext.bgpvpn_port_detach(None, dummy_detach11)

        self.assertEqual(
            2, self.mocked_bagpipe_agent.do_port_plug_refresh.call_count,
            "Do port unplug must be called twice")

        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.delete_bgpvpn(None, dummy_bgpvpn1)

        self.assertEqual(
            0, self.mocked_bagpipe_agent.do_port_plug_refresh.call_count,
            "Do port plug refresh musn't be called")

    def test_bgpvpn_attach_single_port_l3_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        self._check_network_info(base.NETWORK1['id'],
                                 1,
                                 bgpvpn_const.BGPVPN_L3,
                                 base.BGPVPN_L3_RT100)

        # Verify build callback attachments
        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])

        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **self._get_vpn_info(b_const.IPVPN,
                                     local_port['ipvpnif'],
                                     base.BGPVPN_L3_RT100)
            ),
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

    def test_bgpvpn_attach_single_port_l3_bgpvpn_export_only(self):
        bgpvpn_l3_export = copy.copy(base.BGPVPN_L3_RT100)
        bgpvpn_l3_export['import_rt'] = []

        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=bgpvpn_l3_export).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        self._check_network_info(base.NETWORK1['id'],
                                 1,
                                 bgpvpn_const.BGPVPN_L3,
                                 bgpvpn_l3_export)

        # Verify build callback attachments
        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])

        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **self._get_vpn_info(b_const.IPVPN,
                                     local_port['ipvpnif'],
                                     bgpvpn_l3_export)
            ),
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

    def test_bgpvpn_attach_single_port_multiple_bgpvpns(self):
        ipvpn_rts = ({k: base.BGPVPN_L3_RT100[k] + base.BGPVPN_L3_RT200[k]
                      for k in b_const.RT_TYPES})

        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=ipvpn_rts).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        self._check_network_info(base.NETWORK1['id'],
                                 1,
                                 bgpvpn_const.BGPVPN_L3,
                                 ipvpn_rts)

        # Verify build callback attachments
        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])

        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **self._get_vpn_info(b_const.IPVPN,
                                     local_port['ipvpnif'],
                                     ipvpn_rts)
            ),
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

    def test_bgpvpn_attach_multiple_ports_same_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))
        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port11))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])]
        )

        self._check_network_info(base.NETWORK1['id'],
                                 2,
                                 bgpvpn_const.BGPVPN_L3,
                                 base.BGPVPN_L3_RT100)

        # Verify build callback attachments
        for port in [base.PORT10, base.PORT10]:
            local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                       base.NETWORK1['id'],
                                                       port['id'])

            self.assertEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ip_address=port['ip_address'],
                    mac_address=port['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.IPVPN,
                                         local_port['ipvpnif'],
                                         base.BGPVPN_L3_RT100)
                ),
                self.agent_ext.build_bgpvpn_attach_info(port['id'])
            )

    def test_bgpvpn_attach_multiple_ports_different_bgpvpns(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT200).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))
        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port20))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT20['id'])]
        )

        for network_id, bgpvpn_rts in [
                (base.NETWORK1['id'], base.BGPVPN_L3_RT100),
                (base.NETWORK2['id'], base.BGPVPN_L3_RT200)]:
            self._check_network_info(network_id,
                                     1,
                                     bgpvpn_const.BGPVPN_L3,
                                     bgpvpn_rts)

        # Verify build callback attachments
        for port, network, rts in [(base.PORT10, base.NETWORK1,
                                    base.BGPVPN_L3_RT100),
                                   (base.PORT20, base.NETWORK2,
                                    base.BGPVPN_L3_RT200)]:
            local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                       network['id'],
                                                       port['id'])

            self.assertEqual(
                dict(
                    network_id=network['id'],
                    ip_address=port['ip_address'],
                    mac_address=port['mac_address'],
                    gateway_ip=network['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.IPVPN,
                                         local_port['ipvpnif'],
                                         rts)
                ),
                self.agent_ext.build_bgpvpn_attach_info(port['id'])
            )

    def test_bgpvpn_detach_single_port_l3_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        self.agent_ext.bgpvpn_port_attach(None, dummy_port10)

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.bgpvpn_port_detach(None, dummy_detach10)

        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])
        detach_info = {
            b_const.IPVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info)]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 0)

        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)

    def test_bgpvpn_detach_multiple_ports_same_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach11 = dict(id=base.PORT11['id'],
                              network_id=base.NETWORK1['id'])

        # Attach 2 ports on BGP VPN L3 1
        self.agent_ext.bgpvpn_port_attach(None, dummy_port10)
        self.agent_ext.bgpvpn_port_attach(None, dummy_port11)

        # Detach remaining ports from BGP VPN L3 1
        self.agent_ext.bgpvpn_port_detach(None, dummy_detach10)
        self.agent_ext.bgpvpn_port_detach(None, dummy_detach11)

        local_port10 = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                     base.NETWORK1['id'],
                                                     base.PORT10['id'])
        detach_info10 = {
            b_const.IPVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port10['linuxif'])
            }
        }

        local_port11 = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                     base.NETWORK1['id'],
                                                     base.PORT11['id'])
        detach_info11 = {
            b_const.IPVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT11['ip_address'],
                'mac_address': base.PORT11['mac_address'],
                'local_port': dict(linuxif=local_port11['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info10),
             mock.call(base.PORT11['id'], detach_info11)]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 0)

        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)

    def test_bgpvpn_detach_multiple_ports_different_bgpvpns(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT200).__dict__
        dummy_detach20 = dict(id=base.PORT20['id'],
                              network_id=base.NETWORK2['id'])

        self.agent_ext.bgpvpn_port_attach(None, dummy_port10)
        self.agent_ext.bgpvpn_port_attach(None, dummy_port20)

        # Detach all ports from L3 BGP VPNs
        self.agent_ext.bgpvpn_port_detach(None, dummy_detach10)
        self.agent_ext.bgpvpn_port_detach(None, dummy_detach20)

        local_port10 = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                     base.NETWORK1['id'],
                                                     base.PORT10['id'])
        detach_info10 = {
            b_const.IPVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port10['linuxif'])
            }
        }

        local_port20 = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                     base.NETWORK2['id'],
                                                     base.PORT20['id'])
        detach_info20 = {
            b_const.IPVPN: {
                'network_id': base.NETWORK2['id'],
                'ip_address': base.PORT20['ip_address'],
                'mac_address': base.PORT20['mac_address'],
                'local_port': dict(linuxif=local_port20['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info10),
             mock.call(base.PORT20['id'], detach_info20)]
        )

        # Verify attachments list consistency
        for network_id in [base.NETWORK1['id'], base.NETWORK2['id']]:
            self._check_network_info(network_id, 0)

        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)


class TestOVSAgentExtension(base.BaseTestOVSAgentExtension,
                            TestBgpvpnAgentExtensionMixin):

    agent_extension_class = bagpipe_agt_ext.BagpipeBgpvpnAgentExtension

    # Test fallback and ARP gateway voodoo
    def test_fallback(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent_ext.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestOVSAgentExtension,
                  self).test_update_bgpvpn_already_plugged_ports()

            port10_with_gw_mac = base.DummyPort(base.NETWORK1,
                                                base.PORT10).__dict__
            port10_with_gw_mac.update({'gateway_mac': GW_MAC,
                                       'l3vpn': base.BGPVPN_L3_RT100})

            self.agent_ext.bgpvpn_port_attach(None,
                                              copy.copy(port10_with_gw_mac))

            fallback = {'dst_mac': GW_MAC,
                        'ovs_port_number': base.PATCH_MPLS_TO_INT_OFPORT,
                        'src_mac': '00:00:5e:2a:10:00'}

            local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                       base.NETWORK1['id'],
                                                       base.PORT10['id'])

            self.assertEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ip_address=base.PORT10['ip_address'],
                    mac_address=base.PORT10['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.IPVPN,
                                         local_port['ipvpnif'],
                                         base.BGPVPN_L3_RT100,
                                         fallback=fallback)
                ),
                self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
            )

    def test_gateway_arp_voodoo(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent_ext.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent_ext.int_br,
                                  'add_flow') as add_flow, \
                mock.patch.object(self.agent_ext.tun_br,
                                  'delete_flows') as tun_delete_flows,\
                mock.patch.object(self.agent_ext.int_br,
                                  'delete_flows') as int_delete_flows:
            super(TestOVSAgentExtension,
                  self).test_update_bgpvpn_already_plugged_ports()

            port10_with_gw_mac = base.DummyPort(base.NETWORK1,
                                                base.PORT10).__dict__
            port10_with_gw_mac.update({'gateway_mac': GW_MAC,
                                       'l3vpn': base.BGPVPN_L3_RT100})

            self.agent_ext.bgpvpn_port_attach(None,
                                              copy.copy(port10_with_gw_mac))

            self.assertEqual(2, add_flow.call_count)

            add_flow.assert_has_calls([
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x2,
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC,
                          arp_spa='10.0.0.1',
                          actions="drop"),
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x01,
                          dl_src=GW_MAC,
                          arp_spa='10.0.0.1',
                          arp_sha=GW_MAC,
                          actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                          )
            ])

            self.agent_ext.bgpvpn_port_detach(
                None, base.DummyPort(base.NETWORK1, base.PORT10).__dict__
            )

            self.assertEqual(0, tun_delete_flows.call_count)
            self.assertEqual(0, int_delete_flows.call_count)

            self.agent_ext.bgpvpn_port_detach(
                None, base.DummyPort(base.NETWORK1, base.PORT11).__dict__
            )
            self.assertEqual(1, tun_delete_flows.call_count)
            self.assertEqual(1, int_delete_flows.call_count)

            tun_delete_flows.assert_has_calls([
                mock.call(table=mock.ANY,
                          priority=2,
                          strict=True,
                          proto='arp',
                          arp_op=0x01,
                          arp_tpa='10.0.0.1',
                          dl_vlan=mock.ANY,
                          )])
            int_delete_flows.assert_has_calls([
                mock.call(table=mock.ANY,
                          proto='arp',
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC),
            ])

    def test_gateway_arp_voodoo_update_bgpvpn_after_plug(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent_ext.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent_ext.int_br,
                                  'add_flow') as add_flow:
            port10 = base.DummyPort(base.NETWORK1, base.PORT10).__dict__
            port10.update({'gateway_mac': GW_MAC})

            self.agent_ext.bgpvpn_port_attach(None, copy.copy(port10))

            dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                             l3vpn=base.BGPVPN_L3_RT100,
                                             gateway_mac=GW_MAC).__dict__

            self.agent_ext.update_bgpvpn(None, copy.copy(dummy_bgpvpn1))

            self.assertEqual(2, add_flow.call_count)

            add_flow.assert_has_calls([
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x2,
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC,
                          arp_spa='10.0.0.1',
                          actions="drop"),
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x01,
                          dl_src=GW_MAC,
                          arp_spa='10.0.0.1',
                          arp_sha=GW_MAC,
                          actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                          )
            ])

            fallback = {'dst_mac': GW_MAC,
                        'ovs_port_number': base.PATCH_MPLS_TO_INT_OFPORT,
                        'src_mac': '00:00:5e:2a:10:00'}

            self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
                [mock.call(base.PORT10['id'])]
            )

            self.agent_ext.delete_bgpvpn(None, copy.copy(dummy_bgpvpn1))

            add_flow.reset_mock()
            self.assertEqual(0, add_flow.call_count)

            self.agent_ext.update_bgpvpn(None, copy.copy(dummy_bgpvpn1))

            self.assertEqual(2, add_flow.call_count)

            local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                       base.NETWORK1['id'],
                                                       base.PORT10['id'])

            self.assertEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ip_address=base.PORT10['ip_address'],
                    mac_address=base.PORT10['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.IPVPN,
                                         local_port['ipvpnif'],
                                         base.BGPVPN_L3_RT100,
                                         fallback=fallback)
                ),
                self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
            )

    def test_gateway_plug_before_update(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent_ext.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10]), \
                mock.patch.object(self.agent_ext.int_br,
                                  'add_flow') as add_flow:
            port10 = base.DummyPort(base.NETWORK1, base.PORT10).__dict__

            self.agent_ext.bgpvpn_port_attach(None, copy.copy(port10))

            dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                             l3vpn=base.BGPVPN_L3_RT100,
                                             gateway_mac=GW_MAC).__dict__

            update_delete_rpc(dummy_bgpvpn1)

            self.agent_ext.update_bgpvpn(None, copy.copy(dummy_bgpvpn1))

            self.assertEqual(2, add_flow.call_count)

            add_flow.assert_has_calls([
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x2,
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC,
                          arp_spa='10.0.0.1',
                          actions="drop"),
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x01,
                          dl_src=GW_MAC,
                          arp_spa='10.0.0.1',
                          arp_sha=GW_MAC,
                          actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                          )
            ])

            fallback = {'dst_mac': GW_MAC,
                        'ovs_port_number': base.PATCH_MPLS_TO_INT_OFPORT,
                        'src_mac': '00:00:5e:2a:10:00'}

            self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
                [mock.call(base.PORT10['id'])]
            )

            self.agent_ext.delete_bgpvpn(None, copy.copy(dummy_bgpvpn1))

            add_flow.reset_mock()
            self.assertEqual(0, add_flow.call_count)

            self.agent_ext.update_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(2, add_flow.call_count)

            local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                       base.NETWORK1['id'],
                                                       base.PORT10['id'])

            self.assertEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ip_address=base.PORT10['ip_address'],
                    mac_address=base.PORT10['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=local_port['linuxif']),
                    **self._get_vpn_info(b_const.IPVPN,
                                         local_port['ipvpnif'],
                                         base.BGPVPN_L3_RT100,
                                         fallback=fallback)
                ),
                self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
            )

    def test_evpn_no_gateway_arp_voodoo(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent_ext.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent_ext.int_br,
                                  'add_flow') as add_flow, \
                mock.patch.object(self.agent_ext.int_br,
                                  'delete_flows') as delete_flows:

            port10_with_gw_mac = base.DummyPort(base.NETWORK1,
                                                base.PORT10).__dict__
            port10_with_gw_mac.update({'gateway_mac': GW_MAC})
            port10_with_gw_mac.update(
                base.DummyBGPVPN(base.NETWORK1,
                                 l2vpn=base.BGPVPN_L2_RT10).__dict__)

            self.agent_ext.bgpvpn_port_attach(None,
                                              copy.copy(port10_with_gw_mac))

            self.assertEqual(0, add_flow.call_count)

            self.agent_ext.bgpvpn_port_detach(
                None, base.DummyPort(base.NETWORK1, base.PORT10).__dict__
            )
            self.assertEqual(0, delete_flows.call_count)

    def test_gateway_arp_voodoo_ovs_restart(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent_ext.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent_ext.int_br,
                                  'add_flow') as add_flow:
            port10 = base.DummyPort(base.NETWORK1, base.PORT10).__dict__
            port10.update({'gateway_mac': GW_MAC})

            self.agent_ext.bgpvpn_port_attach(None, copy.copy(port10))

            dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                             l3vpn=base.BGPVPN_L3_RT100,
                                             gateway_mac=GW_MAC).__dict__

            update_delete_rpc(dummy_bgpvpn1)

            self.agent_ext.update_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(2, add_flow.call_count)

            expected_calls = [
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x2,
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC,
                          arp_spa='10.0.0.1',
                          actions="drop"),
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x01,
                          dl_src=GW_MAC,
                          arp_spa='10.0.0.1',
                          arp_sha=GW_MAC,
                          actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                          )
            ]
            add_flow.assert_has_calls(expected_calls)

            add_flow.reset_mock()

            self.agent_ext.ovs_restarted_bgpvpn()

            add_flow.assert_has_calls(expected_calls)


class TestLinuxBridgeAgentExtension(base.BaseTestLinuxBridgeAgentExtension,
                                    TestBgpvpnAgentExtensionMixin):

    agent_extension_class = bagpipe_agt_ext.BagpipeBgpvpnAgentExtension

    def _fake_handle_port0(self):
        self.agent_ext.handle_port(None, {
            'port_id': base.PORT10['id'],
            'network_id': base.NETWORK1['id'],
            'segmentation_id': 411,
            'network_type': 'vxlan'
            })

    def test_update_bgpvpn_different_vpn_types(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l2vpn=base.BGPVPN_L2_RT10,
                                         l3vpn=base.BGPVPN_L3_RT100).__dict__

        update_delete_rpc(dummy_bgpvpn1)

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))

        self.agent_ext.update_bgpvpn(None, dummy_bgpvpn1)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        for bgpvpn_type, bgpvpn_rts in [
                (bgpvpn_const.BGPVPN_L2, base.BGPVPN_L2_RT10),
                (bgpvpn_const.BGPVPN_L3, base.BGPVPN_L3_RT100)]:
            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     bgpvpn_type,
                                     bgpvpn_rts)

        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])

        # no handle_port yet => no EVPN yet

        vpns_info = self._get_vpn_info(b_const.IPVPN,
                                       local_port['ipvpnif'],
                                       base.BGPVPN_L3_RT100)

        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **vpns_info
            ),
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

        self._fake_handle_port0()

        # Verify build callback attachments

        vpns_info.update(self._get_vpn_info(b_const.EVPN,
                                            local_port['evpnif'],
                                            base.BGPVPN_L2_RT10,
                                            more_vpn_info={'vni': 411}))

        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **vpns_info
            ),
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

    def test_bgpvpn_attach_single_port_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        self._check_network_info(base.NETWORK1['id'],
                                 1,
                                 bgpvpn_const.BGPVPN_L2,
                                 base.BGPVPN_L2_RT10)

        # before handle_port is called, the info returned is empty
        self.assertEqual(
            {},
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

        # fake handle_port call
        self._fake_handle_port0()

        # now we can check what the build info is...

        # Verify build callback attachments
        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L2,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])

        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **self._get_vpn_info(b_const.EVPN,
                                     local_port['evpnif'],
                                     base.BGPVPN_L2_RT10,
                                     more_vpn_info={'vni': 411})
            ),
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

    def test_bgpvpn_attach_same_port_different_bgpvpn_types(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__
        dummy_port10bis = base.DummyPort(base.NETWORK1, base.PORT10,
                                         bgpvpn_port=True,
                                         evpn=base.BGPVPN_L2_RT10,
                                         ipvpn=base.BGPVPN_L3_RT100).__dict__

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        self._check_network_info(base.NETWORK1['id'],
                                 1,
                                 bgpvpn_const.BGPVPN_L2,
                                 base.BGPVPN_L2_RT10)

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.bgpvpn_port_attach(None, copy.copy(dummy_port10bis))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        for bgpvpn_type, bgpvpn_rts in [
                (bgpvpn_const.BGPVPN_L2, base.BGPVPN_L2_RT10),
                (bgpvpn_const.BGPVPN_L3, base.BGPVPN_L3_RT100)]:
            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     bgpvpn_type,
                                     bgpvpn_rts)

        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])

        # no handle_port received => no EVPN yet

        vpns_info = self._get_vpn_info(b_const.IPVPN,
                                       local_port['ipvpnif'],
                                       base.BGPVPN_L3_RT100)

        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **vpns_info
            ),
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

        # fake handle_port call
        self._fake_handle_port0()

        # Verify build callback attachments

        vpns_info.update(self._get_vpn_info(b_const.EVPN,
                                            local_port['evpnif'],
                                            base.BGPVPN_L2_RT10,
                                            more_vpn_info={'vni': 411}))
        self.assertEqual(
            dict(
                network_id=base.NETWORK1['id'],
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                gateway_ip=base.NETWORK1['gateway_ip'],
                local_port=dict(linuxif=local_port['linuxif']),
                **vpns_info
            ),
            self.agent_ext.build_bgpvpn_attach_info(base.PORT10['id'])
        )

    def test_bgpvpn_detach_single_port_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        self.agent_ext.bgpvpn_port_attach(None, dummy_port10)

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.bgpvpn_port_detach(None, dummy_detach10)

        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L2,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])
        detach_info = {
            b_const.EVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info)]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 0)

        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)

    def test_bgpvpn_detach_single_port_multiple_bgpvpns(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        self.agent_ext.bgpvpn_port_attach(None, dummy_port10)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        self.mocked_bagpipe_agent.reset_mock()

        self.agent_ext.bgpvpn_port_detach(None, dummy_detach10)

        local_port = self._get_expected_local_port(bgpvpn_const.BGPVPN_L3,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'])
        detach_info = {
            b_const.EVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port['linuxif'])
            },
            b_const.IPVPN: {
                'network_id': base.NETWORK1['id'],
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': dict(linuxif=local_port['linuxif'])
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info)]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 0)

        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered attachments list must be empty: %s" %
                         self.agent_ext.networks_info)



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import eventlet
eventlet.monkey_patch()

from oslo_config import cfg

from oslo_concurrency import lockutils

from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from oslo_service import service

from networking_bagpipe.agent import agent_base_info
from networking_bagpipe.agent.common import constants as b_const

from networking_bagpipe.agent import bagpipe_bgp_agent

from networking_bagpipe.driver.type_route_target import TYPE_ROUTE_TARGET

from networking_bagpipe.rpc import agent as bagpipe_rpc
from networking_bagpipe.rpc.client import topics_BAGPIPE

from neutron.common import topics

from neutron_lib.agent import l2_extension

from neutron.common import config as common_config

from neutron_lib import constants as n_const
from neutron_lib.utils import helpers

from neutron.plugins.ml2.drivers.agent import _common_agent as ca
from neutron.plugins.ml2.drivers.linuxbridge.agent.linuxbridge_neutron_agent \
    import LinuxBridgeManager

LOG = logging.getLogger(__name__)

BAGPIPE_L2_SERVICE = 'bagpipe_l2'

LB_BAGPIPE_AGENT_BINARY = 'neutron-bagpipe-linuxbridge-agent'


class LinuxBridgeManagerBaGPipe(LinuxBridgeManager):

    def ensure_physical_in_bridge(self, network_id,
                                  network_type,
                                  physical_network,
                                  segmentation_id):

        if network_type == TYPE_ROUTE_TARGET:
            bridge_name = self.get_bridge_name(network_id)
            return self.ensure_bridge(bridge_name)

        return (super(LinuxBridgeManagerBaGPipe, self)
                .ensure_physical_in_bridge(network_id,
                                           network_type,
                                           physical_network,
                                           segmentation_id))


class BagpipeAgentExtension(l2_extension.L2AgentExtension,
                            agent_base_info.BaseInfoManager,
                            bagpipe_rpc.BaGPipeAgentRpcCallBackMixin):

    def initialize(self, connection, driver_type):

        self.bagpipe_bgp_agent = (
            bagpipe_bgp_agent.BaGPipeBGPAgent.get_instance(
                n_const.AGENT_TYPE_LINUXBRIDGE)
        )

        self._setup_rpc(connection)

        self.bagpipe_bgp_agent.register_build_callback(
            BAGPIPE_L2_SERVICE,
            self.build_bagpipe_l2_attach_info)

        self.ports = set()
        self.bagpipe_bgp_agent.register_port_list(BAGPIPE_L2_SERVICE,
                                                  self.ports)

    def _setup_rpc(self, connection):
        connection.create_consumer(topics.get_topic_name(topics.AGENT,
                                                         topics_BAGPIPE,
                                                         topics.UPDATE,
                                                         cfg.CONF.host),
                                   [self], fanout=False)

    def build_bagpipe_l2_attach_info(self, port_id):
        if not self.ports_info.get(port_id):
            LOG.warning("%s service has no PortInfo for port %s",
                        BAGPIPE_L2_SERVICE, port_id)
            return {}

        port_info = self.ports_info[port_id]

        service_info = port_info.network.service_infos
        attach_info = {
            'network_id': port_info.network.id,
            'ip_address': port_info.ip_address,
            'mac_address': port_info.mac_address,
            'gateway_ip': port_info.network.gateway_info.ip,
            'local_port': port_info.local_port,
            b_const.EVPN: {
                'linuxbr': LinuxBridgeManager.get_bridge_name(
                    port_info.network.id
                ),
                b_const.RT_IMPORT: [
                    service_info[b_const.EVPN][b_const.RT_IMPORT]],
                b_const.RT_EXPORT: [
                    service_info[b_const.EVPN][b_const.RT_EXPORT]]
            }
        }

        return attach_info

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-bgp-agent')
    def bagpipe_port_attach(self, context, port_bagpipe_info):
        port_id = port_bagpipe_info.pop('id')
        net_id = port_bagpipe_info.pop('network_id')

        net_info, port_info = (
            self._get_network_port_infos(net_id, port_id)
        )

        # Set IP and MAC adresses in PortInfo
        ip_address = port_bagpipe_info.pop('ip_address')
        mac_address = port_bagpipe_info.pop('mac_address')
        port_info.set_ip_mac_infos(ip_address, mac_address)

        # Set gateway IP address in NetworkInfo
        gateway_info = b_const.GatewayInfo(None,
                                           port_bagpipe_info.pop('gateway_ip'))
        net_info.set_gateway_info(gateway_info)

        port_info.set_local_port(
            LinuxBridgeManager.get_tap_device_name(port_id)
        )

        net_info.add_service_info(port_bagpipe_info)
        self.ports.add(port_id)

        self.bagpipe_bgp_agent.do_port_plug(port_id)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-bgp-agent')
    def bagpipe_port_detach(self, context, port_bagpipe_info):
        port_id = port_bagpipe_info['id']
        net_id = port_bagpipe_info['network_id']

        if port_id not in self.ports_info:
            LOG.warning("%s service inconsistent for port %s",
                        BAGPIPE_L2_SERVICE, port_id)
            return

        LOG.debug("%s service detaching port %s from bagpipe-bgp",
                  BAGPIPE_L2_SERVICE, port_id)
        try:
            port_info = self.ports_info[port_id]

            detach_info = {
                b_const.EVPN: {
                    'network_id': net_id,
                    'ip_address': port_info.ip_address,
                    'mac_address': port_info.mac_address,
                    'local_port': port_info.local_port,
                }
            }

            self._remove_network_port_infos(net_id, port_id)
            self.ports.remove(port_id)

            self.bagpipe_bgp_agent.do_port_plug_refresh(port_id,
                                                        detach_info)
        except bagpipe_bgp_agent.BaGPipeBGPException as e:
            LOG.error("%s service can't detach port from bagpipe-bgp %s",
                      BAGPIPE_L2_SERVICE, str(e))

    def handle_port(self, context, data):
        pass

    def delete_port(self, context, data):
        pass


def main():
    common_config.init(sys.argv[1:])

    common_config.setup_logging()
    try:
        interface_mappings = helpers.parse_mappings(
            cfg.CONF.LINUX_BRIDGE.physical_interface_mappings)
    except ValueError as e:
        LOG.error("Parsing physical_interface_mappings failed: %s. "
                  "Agent terminated!", e)
        sys.exit(1)
    LOG.info("Interface mappings: %s", interface_mappings)

    try:
        bridge_mappings = helpers.parse_mappings(
            cfg.CONF.LINUX_BRIDGE.bridge_mappings)
    except ValueError as e:
        LOG.error("Parsing bridge_mappings failed: %s. "
                  "Agent terminated!", e)
        sys.exit(1)
    LOG.info("Bridge mappings: %s", bridge_mappings)

    manager = LinuxBridgeManagerBaGPipe(bridge_mappings, interface_mappings)

    polling_interval = cfg.CONF.AGENT.polling_interval
    quitting_rpc_timeout = cfg.CONF.AGENT.quitting_rpc_timeout
    agent = ca.CommonAgentLoop(manager, polling_interval, quitting_rpc_timeout,
                               n_const.AGENT_TYPE_LINUXBRIDGE,
                               LB_BAGPIPE_AGENT_BINARY)
    LOG.info("Agent initialized successfully, now running... ")
    launcher = service.launch(cfg.CONF, agent)
    launcher.wait()
